use crate::{subscription::ActiveSubscription, Error};
use futures_util::{SinkExt, StreamExt};
use nostr_rs_types::{
    relayer::{Auth, ROk},
    types::{Addr, SubscriptionId},
    Request, Response,
};
use std::{
    collections::HashMap,
    sync::atomic::{AtomicUsize, Ordering},
};
use tokio::{
    net::TcpStream,
    sync::{
        mpsc::{channel, Receiver, Sender},
        RwLock,
    },
    task::JoinHandle,
};
#[allow(unused_imports)]
use tokio_tungstenite::{accept_async, tungstenite::Message, WebSocketStream};

#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash, Ord, PartialOrd)]
pub struct ConnectionId(usize);

impl Default for ConnectionId {
    fn default() -> Self {
        static NEXT_ID: AtomicUsize = AtomicUsize::new(1);
        Self(NEXT_ID.fetch_add(1, Ordering::SeqCst))
    }
}

impl ConnectionId {
    /// Connection ID for messages from the client pool or any special empty ConnectionId
    #[inline]
    pub fn new_empty() -> Self {
        Self(0)
    }

    /// Check if the connection id is empty
    ///
    /// Empty connection id is used for messages from Client pool to the relayer
    #[inline]
    pub fn is_empty(&self) -> bool {
        self.0 == 0
    }
}

#[derive(Debug)]
/// Relayer connection
///
/// The new connection struct. This struct spawn's a new worker that handles
/// upcoming messages form the client
pub struct Connection {
    conn_id: ConnectionId,
    sender: Sender<Response>,
    subscriptions: RwLock<HashMap<SubscriptionId, Vec<ActiveSubscription>>>,
    handler: Option<JoinHandle<()>>,
}

const MAX_SUBSCRIPTIONS_BUFFER: usize = 100;

impl Drop for Connection {
    fn drop(&mut self) {
        if let Some(handler) = self.handler.take() {
            let _ = handler.abort();
        }
    }
}

impl Connection {
    #[cfg(test)]
    pub fn new_for_test() -> (Self, Receiver<Response>) {
        let (sender, receiver) = channel(MAX_SUBSCRIPTIONS_BUFFER);
        (
            Self {
                conn_id: ConnectionId::default(),
                sender,
                subscriptions: RwLock::new(HashMap::new()),
                handler: None,
            },
            receiver,
        )
    }

    /// Create new connection
    pub async fn new_connection(
        send_message_to_relayer: Sender<(ConnectionId, Request)>,
        disconnection_notify: Option<Sender<ConnectionId>>,
        stream: TcpStream,
    ) -> Result<Self, Error> {
        let websocket = accept_async(stream).await?;
        let conn_id = Default::default();
        let (sender, receiver) = channel(MAX_SUBSCRIPTIONS_BUFFER);
        let _ = sender.send(Auth::default().into()).await;
        Ok(Self {
            conn_id,
            sender,
            subscriptions: RwLock::new(HashMap::new()),
            handler: Some(Self::spawn(
                send_message_to_relayer,
                websocket,
                receiver,
                disconnection_notify,
                conn_id,
            )),
        })
    }

    fn spawn(
        send_message_to_relayer: Sender<(ConnectionId, Request)>,
        websocket: WebSocketStream<TcpStream>,
        mut receiver: Receiver<Response>,
        disconnection_notify: Option<Sender<ConnectionId>>,
        conn_id: ConnectionId,
    ) -> JoinHandle<()> {
        tokio::spawn(async move {
            let mut _subscriptions: HashMap<String, (u128, Receiver<Response>)> = HashMap::new();
            let (mut writer, mut reader) = websocket.split();
            loop {
                tokio::select! {
                    Some(msg) = receiver.recv() => {
                        let msg = if let Ok(msg) = serde_json::to_string(&msg) {
                            msg
                        } else {
                            continue;
                        };
                        if let Err(err) =  writer.send(Message::Text(msg)).await {
                            log::error!("Error sending message to client: {}", err);
                            break;
                        }
                    }
                    Some(msg) = reader.next() => {
                        if let Ok(Message::Text(msg)) = msg {
                            let msg: Result<Request, _> = serde_json::from_str(&msg);
                            match msg {
                                Ok(msg) => {
                                    let _ = send_message_to_relayer.send((conn_id, msg)).await;
                                },
                                Err(err) => {
                                    log::error!("Error parsing message from client: {}", err);
                                    let reply: Response = ROk {
                                        id: Addr::default(),
                                        status: false,
                                        message: "Error parsing message".to_owned(),
                                    }.into();
                                    let reply = if let Ok(reply) = serde_json::to_string(&reply) {
                                        reply
                                    } else {
                                        continue;
                                    };
                                    if let Err(err) =  writer.send(Message::Text(reply)).await {
                                        log::error!("Error sending message to client: {}", err);
                                        break;
                                    }
                                }
                            };
                        }
                    }
                    else => {
                        break;
                    }
                }
            }

            if let Some(disconnection_notify) = disconnection_notify {
                let _ = disconnection_notify.try_send(conn_id);
            }
        })
    }

    #[inline]
    /// Sends a message to this connection's websocket
    pub fn send(&self, response: Response) -> Result<(), Error> {
        self.sender
            .try_send(response)
            .map_err(|e| Error::TrySendError(Box::new(e)))
    }

    /// Get the sender for this connection
    pub fn get_sender(&self) -> Sender<Response> {
        self.sender.clone()
    }

    /// Get the connection id for this connection
    pub fn get_conn_id(&self) -> ConnectionId {
        self.conn_id
    }

    /// Create a subscription for this connection
    pub async fn keep_track_subscription(
        &self,
        id: SubscriptionId,
        subscriptions: Vec<ActiveSubscription>,
    ) {
        self.subscriptions.write().await.insert(id, subscriptions);
    }
}
