use nostr_rs_types::types::{Event, Filter, Kind, Tag};
use std::collections::BTreeSet;

/// The subscription keys are used to quickly identify the subscriptions
/// by one or more fields. Think of it like a database index
#[derive(Clone, Debug, Eq, PartialEq, PartialOrd, Ord, Hash)]
pub(crate) enum Key {
    /// Key for the author field
    Author(Vec<u8>, Option<Kind>),
    /// Key for the reference to an event
    RefId(Vec<u8>, Option<Kind>),
    /// Key for the reference to a public key
    RefPublicKey(Vec<u8>, Option<Kind>),
    /// Key for the kind field
    Id(Vec<u8>),
    /// Key for the kind field
    Kind(Kind),
    /// Any value, for a catch all
    AllUpdates,
}

type SortedSet<T> = Option<BTreeSet<T>>;

/// Sorted filter
///
/// This is a pre-processed filter that is optimized for fast lookups.
#[derive(Debug, Clone)]
pub struct SortedFilter {
    ids: SortedSet<Vec<u8>>,
    authors: SortedSet<Vec<u8>>,
    kinds: SortedSet<Kind>,
    references_to_event: SortedSet<Vec<u8>>,
    references_to_public_key: SortedSet<Vec<u8>>,
}

impl From<Filter> for SortedFilter {
    /// Converts a Filter into a SortedFilter.
    ///
    /// SortedFilters have SortedSets that are optimized for fast lookups.
    fn from(filter: Filter) -> Self {
        Self {
            ids: (!filter.ids.is_empty()).then(|| {
                filter
                    .ids
                    .into_iter()
                    .map(|x| x.as_ref().to_vec())
                    .collect()
            }),
            authors: (!filter.authors.is_empty()).then(|| {
                filter
                    .authors
                    .into_iter()
                    .map(|x| x.as_ref().to_vec())
                    .collect()
            }),
            kinds: (!filter.kinds.is_empty()).then(|| filter.kinds.into_iter().collect()),
            references_to_event: (!filter.references_to_event.is_empty()).then(|| {
                filter
                    .references_to_event
                    .into_iter()
                    .map(|x| x.as_ref().to_vec())
                    .collect()
            }),
            references_to_public_key: (!filter.references_to_public_key.is_empty()).then(|| {
                filter
                    .references_to_public_key
                    .into_iter()
                    .map(|x| x.as_ref().to_vec())
                    .collect()
            }),
        }
    }
}

impl SortedFilter {
    /// Get the keys for the filter
    ///
    /// Get the keys or indexes for the filter. This is used to quickly identify potential matches
    /// from a set of event listeners.
    pub fn keys(&self) -> Vec<Key> {
        let authors = self
            .authors
            .as_ref()
            .map_or_else(|| vec![], |x| x.iter().map(|x| x.clone()).collect());

        let ids = self
            .ids
            .as_ref()
            .map_or_else(|| vec![], |x| x.iter().map(|x| x.clone()).collect());

        let references_to_event = self
            .references_to_event
            .as_ref()
            .map_or_else(|| vec![], |x| x.iter().map(|x| x.clone()).collect());

        let references_to_public_key = self
            .references_to_public_key
            .as_ref()
            .map_or_else(|| vec![], |x| x.iter().map(|x| x.clone()).collect());

        let kinds = self
            .kinds
            .as_ref()
            .map_or_else(|| vec![], |x| x.iter().map(|x| *x).collect());

        let kind_option = if kinds.is_empty() {
            vec![None]
        } else {
            kinds.clone().into_iter().map(Some).collect()
        };

        let keys = vec![
            authors
                .into_iter()
                .map(|author| {
                    kind_option
                        .iter()
                        .map(|kind| Key::Author(author.clone(), *kind))
                        .collect::<Vec<_>>()
                })
                .collect::<Vec<_>>(),
            references_to_event
                .into_iter()
                .map(|event| {
                    kind_option
                        .iter()
                        .map(|kind| Key::RefId(event.clone(), *kind))
                        .collect::<Vec<_>>()
                })
                .collect::<Vec<_>>(),
            references_to_public_key
                .into_iter()
                .map(|pub_key| {
                    kind_option
                        .iter()
                        .map(|kind| Key::RefPublicKey(pub_key.clone(), *kind))
                        .collect::<Vec<_>>()
                })
                .collect::<Vec<_>>(),
            vec![kinds
                .into_iter()
                .map(|kind| Key::Kind(kind))
                .collect::<Vec<_>>()],
            vec![ids.into_iter().map(|id| Key::Id(id)).collect::<Vec<_>>()],
        ]
        .concat()
        .concat();

        if keys.is_empty() {
            vec![Key::AllUpdates]
        } else {
            keys
        }
    }

    /// Checks if a given key exists in the sorted set, either as a whole or as a partial match
    #[inline]
    fn has_key_or_partial_match<T: AsRef<[u8]>>(id: T, ids: &SortedSet<Vec<u8>>) -> bool {
        if let Some(ids) = ids.as_ref() {
            let id = id.as_ref().to_vec();
            if ids.contains(&id) {
                return true;
            }

            for len in 4..=id.len() {
                let prev_id = &id[..len].to_vec();
                if ids.contains(prev_id) {
                    return true;
                }
            }
            false
        } else {
            true
        }
    }

    /// Checks if any of key given set exists in the sorted set, either as a whole or as a partial match
    #[inline]
    fn has_any_key_or_partial_match<T: Iterator<Item = Vec<u8>>>(
        id_set: T,
        ids: &SortedSet<Vec<u8>>,
    ) -> bool {
        if ids.is_some() {
            for id in id_set {
                if Self::has_key_or_partial_match(id, ids) {
                    return true;
                }
            }
            false
        } else {
            true
        }
    }

    /// Checks if the event matches the filter
    pub fn check(&self, event: &Event) -> bool {
        self.kinds
            .as_ref()
            .map_or(true, |kinds| kinds.contains(&event.kind()))
            && Self::has_key_or_partial_match(&event.id, &self.ids)
            && Self::has_key_or_partial_match(&event.author(), &self.authors)
            && Self::has_any_key_or_partial_match(
                event.tags().iter().filter_map(|f| match f {
                    Tag::Event(x) => Some(x.id.as_ref().to_vec()),
                    _ => None,
                }),
                &self.references_to_event,
            )
            && Self::has_any_key_or_partial_match(
                event.tags().iter().filter_map(|f| match f {
                    Tag::PubKey(x) => Some(x.id.as_ref().to_vec()),
                    _ => None,
                }),
                &self.references_to_public_key,
            )
    }
}
