use super::filter::{Key, SortedFilter};
use crate::connection::ConnectionId;
use nostr_rs_types::{
    client::Subscribe,
    types::{Event, SubscriptionId},
    Response,
};
use std::{
    collections::{BTreeMap, BTreeSet},
    sync::{
        atomic::{AtomicUsize, Ordering},
        Arc,
    },
};
use tokio::sync::{mpsc::Sender, RwLock};

type SubIdx = (Key, ConnectionId, SubscriptionId);

pub const MIN_PREFIX_MATCH_LEN: usize = 4;

/// Subscription for a connection
///
/// This object is responsible for keeping track of a subscription for a connection
///
/// When dropped their listener will be removed from the subscription manager automatically
#[derive(Clone, Debug)]
pub struct ActiveSubscription {
    conn_id: ConnectionId,
    name: SubscriptionId,
    keys: Vec<Key>,
    manager: Arc<SubscriptionManager>,
}

impl ActiveSubscription {
    fn new(
        conn_id: ConnectionId,
        name: SubscriptionId,
        keys: Vec<Key>,
        manager: Arc<SubscriptionManager>,
    ) -> Self {
        Self {
            conn_id,
            name,
            keys,
            manager,
        }
    }
}

impl Drop for ActiveSubscription {
    /// When the subscription is dropped, it will remove the listener from the
    /// subscription manager
    fn drop(&mut self) {
        let keys = self
            .keys
            .drain(..)
            .map(|key| (key, self.conn_id, self.name.clone()))
            .collect::<Vec<_>>();

        let manager = self.manager.clone();

        tokio::spawn(async move {
            manager.unsubscribe(keys).await;
        });
    }
}

/// Subscription manager
///
/// This object is responsible for letting clients and processes subscribe to
/// events,
#[derive(Debug)]
pub struct SubscriptionManager {
    /// List of subscriptions with their filters and their index.
    ///
    /// A single request may be converted to multiple subscriptions entry as
    /// they are sorted by their index/key
    subscriptions: RwLock<BTreeMap<SubIdx, (Arc<SortedFilter>, Sender<Response>)>>,
    /// Total number of subscribers
    /// A single REQ may have multiple subscriptions
    total_subscribers: AtomicUsize,
    /// Minimum prefix match length
    min_prefix_match_len: usize,
}

impl Default for SubscriptionManager {
    fn default() -> Self {
        Self {
            subscriptions: Default::default(),
            total_subscribers: Default::default(),
            min_prefix_match_len: MIN_PREFIX_MATCH_LEN,
        }
    }
}

impl SubscriptionManager {
    async fn unsubscribe(self: Arc<Self>, keys: Vec<SubIdx>) {
        println!("block");
        let mut subscriptions = self.subscriptions.write().await;
        println!("\tblocked");
        for sub in keys {
            subscriptions.remove(&sub);
        }
        self.total_subscribers.fetch_sub(1, Ordering::Relaxed);
        println!("released");
    }

    fn get_keys_from_event(event: &Event, min_prefix_match_len: usize) -> Vec<Key> {
        let mut subscriptions = vec![];

        let author = event.author().as_ref().to_vec();
        let id = event.id.as_ref().to_vec();

        let len = author.len();
        for i in min_prefix_match_len..len - min_prefix_match_len {
            subscriptions.push(Key::Author(author[..len - i].to_vec(), None));
            subscriptions.push(Key::Author(author[..len - i].to_vec(), Some(event.kind())));
        }

        for t in event.tags() {
            match t {
                nostr_rs_types::types::Tag::Event(ref_event) => {
                    let len = ref_event.id.len();
                    for i in min_prefix_match_len..ref_event.id.len() - min_prefix_match_len {
                        subscriptions.push(Key::RefId(ref_event.id[..len - i].to_vec(), None));
                        subscriptions.push(Key::RefId(
                            ref_event.id[..len - i].to_vec(),
                            Some(event.kind()),
                        ));
                    }
                }
                nostr_rs_types::types::Tag::PubKey(ref_pub_key) => {
                    let len = ref_pub_key.id.len();
                    for i in min_prefix_match_len..len - min_prefix_match_len {
                        subscriptions.push(Key::RefId(ref_pub_key.id[..len - i].to_vec(), None));
                        subscriptions.push(Key::RefId(
                            ref_pub_key.id[..len - i].to_vec(),
                            Some(event.kind()),
                        ));
                    }
                }
                _ => {}
            }
        }

        let len = id.len();
        for i in min_prefix_match_len..len - min_prefix_match_len {
            subscriptions.push(Key::Id(id[..len - i].to_vec()));
        }

        subscriptions.push(Key::Kind(event.kind().into()));
        subscriptions.push(Key::AllUpdates);

        subscriptions
    }

    /// Get the number of subscribers
    pub fn total_subscribers(self: &Arc<Self>) -> usize {
        self.total_subscribers.load(Ordering::Relaxed)
    }

    /// Subscribe to a future events
    ///
    /// This will add a new subscription to the subscription manager with a
    /// given conn_id, sender and a vector of filters.
    pub async fn subscribe(
        self: &Arc<Self>,
        conn_id: ConnectionId,
        sender: Sender<Response>,
        request: Subscribe,
    ) -> Vec<ActiveSubscription> {
        let name = request.subscription_id;
        let mut subscriptions = self.subscriptions.write().await;
        let subscriptions = request
            .filters
            .into_iter()
            .map(|filter| {
                let filter = Arc::new(SortedFilter::from(filter));
                let subscription =
                    ActiveSubscription::new(conn_id, name.clone(), filter.keys(), self.clone());
                for key in subscription.keys.iter() {
                    subscriptions.insert(
                        (key.clone(), conn_id, name.clone()),
                        (filter.clone(), sender.clone()),
                    );
                }
                subscription
            })
            .collect::<Vec<_>>();
        self.total_subscribers
            .fetch_add(subscriptions.len(), Ordering::Relaxed);
        subscriptions
    }

    /// Publish an event to all subscribers
    pub fn broadcast(self: &Arc<Self>, event: Event) {
        let this = self.clone();
        tokio::spawn(async move {
            let subscriptions = this.subscriptions.read().await;
            let subs = Self::get_keys_from_event(&event, this.min_prefix_match_len);
            let mut deliverded = BTreeSet::new();

            for sub in subs {
                for ((sub_type, client, name), (filter, sender)) in subscriptions.range(
                    &(
                        sub.clone(),
                        ConnectionId::new_empty(),
                        SubscriptionId::empty(),
                    )..,
                ) {
                    if sub_type != &sub {
                        break;
                    }

                    if deliverded.contains(client) || !filter.check(&event) {
                        continue;
                    }
                    println!("send");

                    let _ = sender.try_send(Response::Event((name, &event).into()));
                    deliverded.insert(client.clone());
                }
            }
        });
    }
}
